import subprocess
from os import PathLike
from pathlib import Path


def get_changed_files(
    base: PathLike[str] = Path("."), *, before: str | None = "main"
) -> set[Path]:
    command: list = [
        "git",
        "-C",
        base,
        "diff",
        "--name-only",
        "--diff-filter=dt",
        before,
    ]
    result = subprocess.run(command, capture_output=True, check=True)
    return {Path(p) for p in result.stdout.decode().strip().split()}



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import pytest

from .diff import get_changed_files
from .selector import Resolver, Selector

POST_REPORT_KEY: pytest.StashKey[str] = pytest.StashKey()
REVISION_KEY: pytest.StashKey[str] = pytest.StashKey()
USE_SHORT_CIRCUIT_KEY: pytest.StashKey[bool] = pytest.StashKey()
INCLUDED_MODULES_KEY: pytest.StashKey[set[str]] = pytest.StashKey()


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("collect", "collection")
    group.addoption("--gitscope", help="Select tests based on git revision")
    group.addoption(
        "--gitscope-no-short-circuits",
        action="store_true",
        default=False,
        help="Do not use the short circuits",
    )

    group.addoption(
        "--gitscope-include-module",
        action="append",
        metavar="MODULE",
        help="Include tests that depends on this module and its submodules",
    )

    # Let user register custom short circuit files
    parser.addini(
        "gitscope_short_circuits",
        "list of (relative) glob-style paths to be used for short circuit.",
        type="paths",
        default=default_short_circuit_files(),
    )


def pytest_configure(config: pytest.Config):
    if rev := config.getoption("--gitscope"):
        config.stash[REVISION_KEY] = rev
        config.stash[USE_SHORT_CIRCUIT_KEY] = not config.getoption(
            "--gitscope-no-short-circuits"
        )
        config.stash[INCLUDED_MODULES_KEY] = set(
            config.getoption("--gitscope-include-module") or []
        )


def pytest_report_header(config: pytest.Config, start_path: Any) -> str | None:
    if rev := config.stash.get(REVISION_KEY, None):
        return f"gitscope: Analyzing changes from {rev}"
    else:
        return None


@pytest.hookimpl(tryfirst=True)
def pytest_collection_modifyitems(
    session: pytest.Session, config: pytest.Config, items: list[pytest.Item]
) -> None:
    if not items:
        return
    rev = config.stash.get(REVISION_KEY, None)
    if rev is None:
        return

    root = session.startpath
    changed_files = get_changed_files(root, before=rev)

    if not changed_files:
        return

    # Track changes of short circuit files. if a short circuit is changed, then short circuit the whole thing
    # This kind of file usually declares dependencies that are difficult to inspect
    if config.stash.get(USE_SHORT_CIRCUIT_KEY, True) and (
        short_circuit_files := unfold_files(
            root, config.getini("gitscope_short_circuits")
        )
    ):
        if matched_short_circuit_files := changed_files & short_circuit_files:
            # A file that may declare some external dependencies have been changed.
            # it safer to not try to filter
            config.stash[POST_REPORT_KEY] = (
                "The pytest-gitscope plugin won't try to deselect some tests, "
                f"because these files ({', '.join(sorted(map(str, matched_short_circuit_files)))}) have been changed since {rev}"
            )
            return

    # Track changes of conftest.py files. if a conftest.py is changed, then short circuit the whole thing
    changed_conftest_files = {
        changed_file
        for changed_file in changed_files
        if changed_file.name in ["conftest.py"]
    }
    if changed_conftest_files:
        # Some conftest.py have been changed.
        # it safer to not try to filter
        config.stash[POST_REPORT_KEY] = (
            "The pytest-gitscope plugin won't try to deselect some tests, "
            f"because it cannot detect changes introduced into ({', '.join(sorted(map(str, changed_conftest_files)))}) since {rev}"
        )
        return

    selector = Selector(
        changed_files=changed_files,
        resolver=Resolver.from_modules(root=root, modules=sys.modules),
        included_modules=config.stash[INCLUDED_MODULES_KEY],
    )

    # those will be our bases
    test_files = {item.path.relative_to(root) for item in items}
    test_dirs: set[Path] = set()
    for test_file in test_files:
        test_dirs.update(test_file.parents)

    # Track dependencies' changes into conftest.py files. if a conftest.py is affected by a dependency change, then short circuit the whole thing
    conftest_files = {
        conftest_file
        for test_dir in test_dirs
        if (conftest_file := test_dir / "conftest.py") and conftest_file.exists()
    }

    affected_conftest_files = selector.select_files(target_files=conftest_files)
    if affected_conftest_files:
        # Some conftest.py files have been affected by changes.
        # Because they do declare fixtures, it is safer to not try to filter
        config.stash[POST_REPORT_KEY] = (
            "The pytest-gitscope plugin won't try to deselect some tests, "
            f"because file ({', '.join(sorted(map(str, affected_conftest_files)))}) have been affected by dependency changes since {rev}"
        )
        return

    affected_test_files = selector.select_files(target_files=test_files)

    remaining = []
    deselected = []
    for item in items:
        if item.path.relative_to(root) in affected_test_files:
            remaining.append(item)
        else:
            deselected.append(item)
    if deselected:
        config.hook.pytest_deselected(items=deselected)
        items[:] = remaining
        config.stash[POST_REPORT_KEY] = (
            "Some tests have been deselected by pytest-gitscope plugin, "
            f"because they have not been affected by the changes from {rev}"
        )


def pytest_report_collectionfinish(
    config: pytest.Config, start_path: Any, startdir: Any, items: Any
) -> str | list[str]:
    if data := config.stash.get(POST_REPORT_KEY, default=None):
        return data
    return []


def default_short_circuit_files():
    return {
        Path("pyproject.toml"),
        Path("requirements.txt"),
        Path("poetry.lock"),
        Path("uv.lock"),
        Path("pylock.toml"),
        Path("Pipfile.lock"),
        Path("Pipfile"),
        Path("pdm.lock"),
        Path("setup.cfg"),
        Path("setup.py"),
        Path("requirements.in"),
        Path("pytest.ini"),
    }


def unfold_files(root: Path, custom_paths: list[Path] | None) -> set[Path]:
    unfolded_files: set[Path] = set()
    if custom_paths:
        for custom_path in custom_paths:
            if root in custom_path.parents:
                custom_path = custom_path.relative_to(root)
            if "*" in str(custom_path):
                unfolded_files.update(Path().glob(str(custom_path)))
            else:
                unfolded_files.add(custom_path)
    return unfolded_files



from __future__ import annotations

import ast
from dataclasses import dataclass, field
from functools import cache
from importlib.util import find_spec
from pathlib import Path
from types import ModuleType
from typing import NamedTuple, Self, TypeAlias

Name: TypeAlias = str


class Module(NamedTuple):
    name: Name
    file: Path | None


@dataclass
class Resolver:
    root: Path
    by_names: dict[Name, Path | None] = field(default_factory=dict)
    by_files: dict[Path, Name] = field(default_factory=dict)

    @classmethod
    def from_modules(cls, root: Path, modules: dict[str, ModuleType]) -> Self:
        by_names: dict[Name, Path | None] = {}
        by_files: dict[Path, Name] = {}
        for name, module in modules.items():
            if (
                (filepath := getattr(module, "__file__", None))
                and (file := Path(filepath))
                and root in file.parents
            ):
                file = file.relative_to(root)
                by_names[name] = file
                by_files[file] = name
            else:
                by_names[name] = None
        return cls(root=root, by_names=by_names, by_files=by_files)

    def __post_init__(self) -> None:
        self.infer_dependencies = cache(self.infer_dependencies)  # type: ignore[method-assign]
        self.get_module = cache(self.get_module)  # type: ignore[method-assign]
        self.venv_root = self.root / ".venv"

    def get_module_by_file(self, file: Path) -> Module | None:
        if name := self.by_files.get(file):
            return Module(name, file)
        else:
            return None

    def get_module(self, name: str) -> Module | None:
        try:
            file = self.by_names[name]
            return Module(name, file)
        except KeyError:
            pass

        # Fetch from find_spec
        try:
            spec = find_spec(name)
        except ModuleNotFoundError:
            return None

        if spec:
            if (
                spec.origin
                and (file := Path(spec.origin))
                and file.suffix in (".py",)  # For now handle only python files
                and (self.root in file.parents)
                and (self.venv_root not in file.parents)  # Eliminate virtualenv folders
            ):
                file = file.relative_to(self.root)
            else:
                file = None
            return Module(name, file)
        return None

    def infer_dependencies(self, mod: Module) -> set[Name]:
        dependency_names: set[str] = set()
        if not mod.file:
            return dependency_names

        try:
            source = mod.file.read_text()
        except Exception as error:
            error.add_note(f"Current file {mod.file} is expected to be a dependency")
            error.add_note(f"Dependency name is {mod.name}")
            raise error

        parts = tuple(mod.name.split("."))
        tree = ast.parse(source, filename=mod.file)
        for node in ast.walk(tree):
            match node:
                case ast.Import(names):
                    for name in names:
                        dependency_names.add(name.name)
                case ast.ImportFrom(None, names, level):
                    assert len(parts) >= level
                    for name in names:
                        dependency_names.add(".".join(parts[-level:] + (name.name,)))
                case ast.ImportFrom(str(module), names, 0):
                    for name in names:
                        dependency_names.add(".".join((module, name.name)))
                case ast.ImportFrom(str(module), names, level):
                    assert len(parts) >= level
                    for name in names:
                        dependency_names.add(
                            ".".join(parts[-level:] + (module, name.name))
                        )

        for dependency_name in list(dependency_names):
            while "." in dependency_name:
                dependency_name, *_ = dependency_name.rpartition(".")
                dependency_names.add(dependency_name)
        return dependency_names

    def match(self, mod: Module, *, files: set[Path], modules: set[Name]) -> bool:
        if mod.name in modules:
            return True
        if mod.file in files:
            return True
        resolved: set[Name] = {mod.name}
        queue = [mod]
        while queue:
            mods, queue = queue, []
            for mod in mods:
                for dependency_name in self.infer_dependencies(mod) - resolved:
                    resolved.add(mod.name)
                    if dependency_name in modules:
                        return True
                    if dependency := self.get_module(name=dependency_name):
                        if dependency.file in files:
                            return True
                        queue.append(dependency)
        return False


@dataclass
class Selector:
    changed_files: set[Path]
    resolver: Resolver
    included_modules: set[Name]

    def select_files(self, target_files: set[Path]) -> set[Path]:
        selection = target_files & self.changed_files
        target_files = target_files - selection

        if not target_files:
            # we already took everything
            return selection

        for target_file in target_files:
            if mod := self.resolver.get_module_by_file(target_file):
                if self.resolver.match(
                    mod, files=self.changed_files, modules=self.included_modules
                ):
                    selection.add(target_file)
        return selection
